// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: transportinstruction.proto
// Protobuf C++ Version: 7.34.1

#ifndef transportinstruction_2eproto_2epb_2eh
#define transportinstruction_2eproto_2epb_2eh

#include <limits>
#include <string>
#include <type_traits>
#include <utility>

// clang-format off
#include "google/protobuf/runtime_version.h"
#if PROTOBUF_VERSION != 7034001
#error "Protobuf C++ gencode is built with an incompatible version of"
#error "Protobuf C++ headers/runtime. See"
#error "https://protobuf.dev/support/cross-version-runtime-guarantee/#cpp"
#endif
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/arena.h"
#include "google/protobuf/arenastring.h"
#include "google/protobuf/generated_message_tctable_decl.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/metadata_lite.h"
#include "google/protobuf/message_lite.h"
#include "google/protobuf/repeated_field.h"  // IWYU pragma: export
#include "google/protobuf/extension_set.h"  // IWYU pragma: export
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"

#define PROTOBUF_INTERNAL_EXPORT_transportinstruction_2eproto

namespace google {
namespace protobuf {
namespace internal {
template <typename T>
::absl::string_view GetAnyMessageName();
}  // namespace internal
}  // namespace protobuf
}  // namespace google

// Internal implementation detail -- do not use these members.
struct TableStruct_transportinstruction_2eproto {
  static const ::uint32_t offsets[];
};
namespace TransportBuffers {
class Instruction;
struct InstructionDefaultTypeInternal;
extern InstructionDefaultTypeInternal _Instruction_default_instance_;
extern const ::google::protobuf::internal::ClassDataLite<29> Instruction_class_data_;
}  // namespace TransportBuffers
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google

namespace TransportBuffers {

// ===================================================================


// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED Instruction final : public ::google::protobuf::MessageLite
/* @@protoc_insertion_point(class_definition:TransportBuffers.Instruction) */ {
 public:
  inline Instruction() : Instruction(nullptr) {}
  ~Instruction() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(Instruction* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(Instruction));
  }
#endif

  template <typename = void>
  explicit constexpr Instruction(::google::protobuf::internal::ConstantInitialized);

  inline Instruction(const Instruction& from) : Instruction(nullptr, from) {}
  inline Instruction(Instruction&& from) noexcept
      : Instruction(nullptr, ::std::move(from)) {}
  inline Instruction& operator=(const Instruction& from) {
    CopyFrom(from);
    return *this;
  }
  inline Instruction& operator=(Instruction&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::std::string& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::std::string>(::google::protobuf::internal::GetEmptyString);
  }
  [[nodiscard]] inline ::std::string* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::std::string>();
  }

  [[nodiscard]] static const Instruction& default_instance() {
    return *reinterpret_cast<const Instruction*>(
        &_Instruction_default_instance_);
  }
  static constexpr int kIndexInFileMessages = 0;
  friend void swap(Instruction& a, Instruction& b) { a.Swap(&b); }
  inline void Swap(Instruction* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(Instruction* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] Instruction* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::MessageLite::DefaultConstruct<Instruction>(arena);
  }
  void CopyFrom(const Instruction& from);
  void MergeFrom(const Instruction& from) { Instruction::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(Instruction* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "TransportBuffers.Instruction"; }

  explicit Instruction(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  Instruction(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Instruction& from);
  Instruction(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, Instruction&& from) noexcept
      : Instruction(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_();

  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kDiffFieldNumber = 6,
    kChaffFieldNumber = 7,
    kOldNumFieldNumber = 2,
    kNewNumFieldNumber = 3,
    kAckNumFieldNumber = 4,
    kThrowawayNumFieldNumber = 5,
    kProtocolVersionFieldNumber = 1,
  };
  // optional bytes diff = 6;
  [[nodiscard]] bool has_diff()
      const;
  void clear_diff() ;
  [[nodiscard]] const ::std::string& diff() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_diff(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_diff();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_diff();
  void set_allocated_diff(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_diff() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_diff(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_diff();

  public:
  // optional bytes chaff = 7;
  [[nodiscard]] bool has_chaff()
      const;
  void clear_chaff() ;
  [[nodiscard]] const ::std::string& chaff() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_chaff(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_chaff();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_chaff();
  void set_allocated_chaff(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_chaff() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_chaff(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_chaff();

  public:
  // optional uint64 old_num = 2;
  [[nodiscard]] bool has_old_num()
      const;
  void clear_old_num() ;
  [[nodiscard]] ::uint64_t old_num() const;
  void set_old_num(::uint64_t value);

  private:
  ::uint64_t _internal_old_num() const;
  void _internal_set_old_num(::uint64_t value);

  public:
  // optional uint64 new_num = 3;
  [[nodiscard]] bool has_new_num()
      const;
  void clear_new_num() ;
  [[nodiscard]] ::uint64_t new_num() const;
  void set_new_num(::uint64_t value);

  private:
  ::uint64_t _internal_new_num() const;
  void _internal_set_new_num(::uint64_t value);

  public:
  // optional uint64 ack_num = 4;
  [[nodiscard]] bool has_ack_num()
      const;
  void clear_ack_num() ;
  [[nodiscard]] ::uint64_t ack_num() const;
  void set_ack_num(::uint64_t value);

  private:
  ::uint64_t _internal_ack_num() const;
  void _internal_set_ack_num(::uint64_t value);

  public:
  // optional uint64 throwaway_num = 5;
  [[nodiscard]] bool has_throwaway_num()
      const;
  void clear_throwaway_num() ;
  [[nodiscard]] ::uint64_t throwaway_num() const;
  void set_throwaway_num(::uint64_t value);

  private:
  ::uint64_t _internal_throwaway_num() const;
  void _internal_set_throwaway_num(::uint64_t value);

  public:
  // optional uint32 protocol_version = 1;
  [[nodiscard]] bool has_protocol_version()
      const;
  void clear_protocol_version() ;
  [[nodiscard]] ::uint32_t protocol_version() const;
  void set_protocol_version(::uint32_t value);

  private:
  ::uint32_t _internal_protocol_version() const;
  void _internal_set_protocol_version(::uint32_t value);

  public:
  // @@protoc_insertion_point(class_scope:TransportBuffers.Instruction)
 private:
  class _Internal;
  friend class ::google::protobuf::internal::TcParser;
  static const ::google::protobuf::internal::TcParseTable<3, 7,
                                   0, 0,
                                   2>
      _table_;

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const Instruction& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::internal::ArenaStringPtr diff_;
    ::google::protobuf::internal::ArenaStringPtr chaff_;
    ::uint64_t old_num_;
    ::uint64_t new_num_;
    ::uint64_t ack_num_;
    ::uint64_t throwaway_num_;
    ::uint32_t protocol_version_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_transportinstruction_2eproto;
};

extern const ::google::protobuf::internal::ClassDataLite<29> Instruction_class_data_;

// ===================================================================




// ===================================================================


#ifdef __GNUC__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-aliasing"
#endif  // __GNUC__
// -------------------------------------------------------------------

// Instruction

// optional uint32 protocol_version = 1;
inline bool Instruction::has_protocol_version() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000040U);
  return value;
}
inline void Instruction::clear_protocol_version() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.protocol_version_ = 0u;
  ClearHasBit(_impl_._has_bits_[0],
                  0x00000040U);
}
inline ::uint32_t Instruction::protocol_version() const {
  // @@protoc_insertion_point(field_get:TransportBuffers.Instruction.protocol_version)
  return _internal_protocol_version();
}
inline void Instruction::set_protocol_version(::uint32_t value) {
  _internal_set_protocol_version(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000040U);
  // @@protoc_insertion_point(field_set:TransportBuffers.Instruction.protocol_version)
}
inline ::uint32_t Instruction::_internal_protocol_version() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.protocol_version_;
}
inline void Instruction::_internal_set_protocol_version(::uint32_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.protocol_version_ = value;
}

// optional uint64 old_num = 2;
inline bool Instruction::has_old_num() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000004U);
  return value;
}
inline void Instruction::clear_old_num() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.old_num_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0],
                  0x00000004U);
}
inline ::uint64_t Instruction::old_num() const {
  // @@protoc_insertion_point(field_get:TransportBuffers.Instruction.old_num)
  return _internal_old_num();
}
inline void Instruction::set_old_num(::uint64_t value) {
  _internal_set_old_num(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  // @@protoc_insertion_point(field_set:TransportBuffers.Instruction.old_num)
}
inline ::uint64_t Instruction::_internal_old_num() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.old_num_;
}
inline void Instruction::_internal_set_old_num(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.old_num_ = value;
}

// optional uint64 new_num = 3;
inline bool Instruction::has_new_num() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000008U);
  return value;
}
inline void Instruction::clear_new_num() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.new_num_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0],
                  0x00000008U);
}
inline ::uint64_t Instruction::new_num() const {
  // @@protoc_insertion_point(field_get:TransportBuffers.Instruction.new_num)
  return _internal_new_num();
}
inline void Instruction::set_new_num(::uint64_t value) {
  _internal_set_new_num(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  // @@protoc_insertion_point(field_set:TransportBuffers.Instruction.new_num)
}
inline ::uint64_t Instruction::_internal_new_num() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.new_num_;
}
inline void Instruction::_internal_set_new_num(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.new_num_ = value;
}

// optional uint64 ack_num = 4;
inline bool Instruction::has_ack_num() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000010U);
  return value;
}
inline void Instruction::clear_ack_num() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.ack_num_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0],
                  0x00000010U);
}
inline ::uint64_t Instruction::ack_num() const {
  // @@protoc_insertion_point(field_get:TransportBuffers.Instruction.ack_num)
  return _internal_ack_num();
}
inline void Instruction::set_ack_num(::uint64_t value) {
  _internal_set_ack_num(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  // @@protoc_insertion_point(field_set:TransportBuffers.Instruction.ack_num)
}
inline ::uint64_t Instruction::_internal_ack_num() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.ack_num_;
}
inline void Instruction::_internal_set_ack_num(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.ack_num_ = value;
}

// optional uint64 throwaway_num = 5;
inline bool Instruction::has_throwaway_num() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000020U);
  return value;
}
inline void Instruction::clear_throwaway_num() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.throwaway_num_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0],
                  0x00000020U);
}
inline ::uint64_t Instruction::throwaway_num() const {
  // @@protoc_insertion_point(field_get:TransportBuffers.Instruction.throwaway_num)
  return _internal_throwaway_num();
}
inline void Instruction::set_throwaway_num(::uint64_t value) {
  _internal_set_throwaway_num(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000020U);
  // @@protoc_insertion_point(field_set:TransportBuffers.Instruction.throwaway_num)
}
inline ::uint64_t Instruction::_internal_throwaway_num() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.throwaway_num_;
}
inline void Instruction::_internal_set_throwaway_num(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.throwaway_num_ = value;
}

// optional bytes diff = 6;
inline bool Instruction::has_diff() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000001U);
  return value;
}
inline void Instruction::clear_diff() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.diff_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0],
                  0x00000001U);
}
inline const ::std::string& Instruction::diff() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:TransportBuffers.Instruction.diff)
  return _internal_diff();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void Instruction::set_diff(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  _impl_.diff_.SetBytes(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:TransportBuffers.Instruction.diff)
}
inline ::std::string* PROTOBUF_NONNULL Instruction::mutable_diff()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  ::std::string* _s = _internal_mutable_diff();
  // @@protoc_insertion_point(field_mutable:TransportBuffers.Instruction.diff)
  return _s;
}
inline const ::std::string& Instruction::_internal_diff() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.diff_.Get();
}
inline void Instruction::_internal_set_diff(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.diff_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL Instruction::_internal_mutable_diff() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.diff_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE Instruction::release_diff() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:TransportBuffers.Instruction.diff)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000001U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  auto* released = _impl_.diff_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.diff_.Set("", GetArena());
  }
  return released;
}
inline void Instruction::set_allocated_diff(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  }
  _impl_.diff_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.diff_.IsDefault()) {
    _impl_.diff_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:TransportBuffers.Instruction.diff)
}

// optional bytes chaff = 7;
inline bool Instruction::has_chaff() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000002U);
  return value;
}
inline void Instruction::clear_chaff() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.chaff_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0],
                  0x00000002U);
}
inline const ::std::string& Instruction::chaff() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:TransportBuffers.Instruction.chaff)
  return _internal_chaff();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void Instruction::set_chaff(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  _impl_.chaff_.SetBytes(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:TransportBuffers.Instruction.chaff)
}
inline ::std::string* PROTOBUF_NONNULL Instruction::mutable_chaff()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::std::string* _s = _internal_mutable_chaff();
  // @@protoc_insertion_point(field_mutable:TransportBuffers.Instruction.chaff)
  return _s;
}
inline const ::std::string& Instruction::_internal_chaff() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.chaff_.Get();
}
inline void Instruction::_internal_set_chaff(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.chaff_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL Instruction::_internal_mutable_chaff() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.chaff_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE Instruction::release_chaff() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:TransportBuffers.Instruction.chaff)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000002U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  auto* released = _impl_.chaff_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.chaff_.Set("", GetArena());
  }
  return released;
}
inline void Instruction::set_allocated_chaff(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }
  _impl_.chaff_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.chaff_.IsDefault()) {
    _impl_.chaff_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:TransportBuffers.Instruction.chaff)
}

#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif  // __GNUC__

// @@protoc_insertion_point(namespace_scope)
}  // namespace TransportBuffers


// @@protoc_insertion_point(global_scope)

#include "google/protobuf/port_undef.inc"
// clang-format on

#endif  // transportinstruction_2eproto_2epb_2eh
